==> ideaIC-201.6668.113-sources.jar: git4idea/ui/branch/dashboard/BranchesDashboardUi.kt <==
// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package git4idea.ui.branch.dashboard

import com.intellij.icons.AllIcons
import com.intellij.ide.CommonActionsManager
import com.intellij.ide.DataManager
import com.intellij.ide.DefaultTreeExpander
import com.intellij.openapi.Disposable
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.keymap.KeymapUtil
import com.intellij.openapi.progress.util.ProgressWindow
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.registry.Registry
import com.intellij.openapi.wm.IdeFocusManager
import com.intellij.ui.IdeBorderFactory.createBorder
import com.intellij.ui.JBColor
import com.intellij.ui.OnePixelSplitter
import com.intellij.ui.ScrollPaneFactory
import com.intellij.ui.SideBorder
import com.intellij.ui.components.panels.NonOpaquePanel
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.JBUI.Panels.simplePanel
import com.intellij.util.ui.StatusText.getDefaultEmptyText
import com.intellij.util.ui.UIUtil
import com.intellij.util.ui.components.BorderLayoutPanel
import com.intellij.util.ui.table.ComponentsListFocusTraversalPolicy
import com.intellij.vcs.log.VcsLogFilterCollection
import com.intellij.vcs.log.data.VcsLogData
import com.intellij.vcs.log.impl.MainVcsLogUiProperties
import com.intellij.vcs.log.impl.VcsLogManager
import com.intellij.vcs.log.impl.VcsLogManager.BaseVcsLogUiFactory
import com.intellij.vcs.log.impl.VcsLogProjectTabsProperties
import com.intellij.vcs.log.impl.VcsLogProjectTabsProperties.MAIN_LOG_ID
import com.intellij.vcs.log.impl.VcsLogUiProperties
import com.intellij.vcs.log.ui.VcsLogColorManager
import com.intellij.vcs.log.ui.VcsLogColorManagerImpl
import com.intellij.vcs.log.ui.VcsLogInternalDataKeys
import com.intellij.vcs.log.ui.VcsLogUiImpl
import com.intellij.vcs.log.ui.filter.VcsLogFilterUiEx
import com.intellij.vcs.log.ui.frame.*
import com.intellij.vcs.log.util.VcsLogUiUtil.isDiffPreviewInEditor
import com.intellij.vcs.log.visible.VisiblePackRefresher
import com.intellij.vcs.log.visible.VisiblePackRefresherImpl
import com.intellij.vcs.log.visible.filters.VcsLogFilterObject
import git4idea.i18n.GitBundle.message
import git4idea.i18n.GitBundleExtensions.messagePointer
import git4idea.ui.branch.dashboard.BranchesDashboardActions.DeleteBranchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.FetchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.GroupByDirectoryAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.NewBranchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ShowBranchDiffAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ShowMyBranchesAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ToggleFavoriteAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.UpdateSelectedBranchAction
import org.jetbrains.annotations.CalledInAwt
import java.awt.Component
import javax.swing.JComponent
import javax.swing.event.TreeSelectionListener

internal class BranchesDashboardUi(project: Project, private val logUi: BranchesVcsLogUi) : Disposable {
  private val uiController = BranchesDashboardController(project, this)

  private val tree = FilteringBranchesTree(project, BranchesTreeComponent(project), uiController)
  private val branchViewSplitter = BranchViewSplitter()
  private val branchesTreePanel = BranchesTreePanel().withBorder(createBorder(JBColor.border(), SideBorder.LEFT))
  private val branchesScrollPane = ScrollPaneFactory.createScrollPane(tree.component, true)
  private val branchesProgressStripe = ProgressStripe(branchesScrollPane, this, ProgressWindow.DEFAULT_PROGRESS_DIALOG_POSTPONE_TIME_MILLIS)
  private val branchesTreeWithLogPanel = simplePanel()
  private val mainPanel = simplePanel().apply { DataManager.registerDataProvider(this, uiController) }
  private val branchesSearchFieldPanel = simplePanel()
  private val branchesSearchField = NonOpaquePanel(tree.installSearchField(JBUI.Borders.emptyLeft(5))).apply(UIUtil::setNotOpaqueRecursively)

  private lateinit var branchesPanelExpandableController: ExpandablePanelController

  private val treeSelectionListener = TreeSelectionListener {
    if (!branchesPanelExpandableController.isExpanded()) return@TreeSelectionListener

    val ui = logUi
    val branchNames = tree.getSelectedBranchNames()
    ui.filterUi.setFilter(if (branchNames.isNotEmpty()) VcsLogFilterObject.fromBranches(branchNames) else null)
  }

  private val BRANCHES_UI_FOCUS_TRAVERSAL_POLICY = object : ComponentsListFocusTraversalPolicy() {
    override fun getOrderedComponents(): List<Component> = listOf(tree.component, logUi.table,
                                                                  logUi.mainFrame.changesBrowser.preferredFocusedComponent,
                                                                  logUi.mainFrame.filterUi.textFilterComponent.textEditor)
  }

  init {
    initMainUi()
    installLogUi()
    updateBranchesTree(true)
  }

  @CalledInAwt
  private fun installLogUi() {
    uiController.registerDataPackListener(logUi.logData)
    uiController.registerLogUiPropertiesListener(logUi.properties)
    branchesSearchField.setVerticalSizeReferent(logUi.toolbar)
    branchViewSplitter.secondComponent = logUi.mainFrame
    val isDiffPreviewInEditor = isDiffPreviewInEditor()
    val diffPreview = logUi.createDiffPreview(isDiffPreviewInEditor)
    if (isDiffPreviewInEditor) {
      mainPanel.add(branchesTreeWithLogPanel)
    }
    else {
      mainPanel.add(DiffPreviewSplitter(diffPreview, logUi.properties, branchesTreeWithLogPanel).mainComponent)
    }
    tree.component.addTreeSelectionListener(treeSelectionListener)
  }

  @CalledInAwt
  private fun disposeBranchesUi() {
    branchViewSplitter.secondComponent.removeAll()
    uiController.removeDataPackListener(logUi.logData)
    uiController.removeLogUiPropertiesListener(logUi.properties)
    tree.component.removeTreeSelectionListener(treeSelectionListener)
  }

  private fun initMainUi() {
    val diffAction = ShowBranchDiffAction()
    diffAction.registerCustomShortcutSet(KeymapUtil.getActiveKeymapShortcuts("Diff.ShowDiff"), branchesTreeWithLogPanel)

    val deleteAction = DeleteBranchAction()
    val shortcuts = KeymapUtil.getActiveKeymapShortcuts("SafeDelete").shortcuts + KeymapUtil.getActiveKeymapShortcuts(
      "EditorDeleteToLineStart").shortcuts
    deleteAction.registerCustomShortcutSet(CustomShortcutSet(*shortcuts), branchesTreeWithLogPanel)

    createFocusFilterFieldAction(branchesSearchField)

    val groupByDirectoryAction = GroupByDirectoryAction(tree)
    val toggleFavoriteAction = ToggleFavoriteAction()
    val fetchAction = FetchAction(this)
    val showMyBranchesAction = ShowMyBranchesAction(uiController)
    val newBranchAction = NewBranchAction()
    val updateSelectedAction = UpdateSelectedBranchAction()
    val defaultTreeExpander = DefaultTreeExpander(tree.component)
    val commonActionsManager = CommonActionsManager.getInstance()
    val expandAllAction = commonActionsManager.createExpandAllHeaderAction(defaultTreeExpander, tree.component)
    val collapseAllAction = commonActionsManager.createCollapseAllHeaderAction(defaultTreeExpander, tree.component)
    val hideBranchesAction = ActionManager.getInstance().getAction("Git.Log.Hide.Branches")

    val group = DefaultActionGroup()
    group.add(hideBranchesAction)
    group.add(Separator())
    group.add(newBranchAction)
    group.add(updateSelectedAction)
    group.add(deleteAction)
    group.add(diffAction)
    group.add(showMyBranchesAction)
    group.add(fetchAction)
    group.add(toggleFavoriteAction)
    group.add(Separator())
    group.add(groupByDirectoryAction)
    group.add(expandAllAction)
    group.add(collapseAllAction)

    val toolbar = ActionManager.getInstance().createActionToolbar("Git.Log.Branches", group, false)
    toolbar.setTargetComponent(branchesTreePanel)

    val branchesButton = ExpandStripeButton(messagePointer("action.Git.Log.Show.Branches.text"), AllIcons.Actions.ArrowExpand)
      .apply {
        border = createBorder(JBColor.border(), SideBorder.RIGHT)
        addActionListener {
          if (logUi.properties.exists(SHOW_GIT_BRANCHES_LOG_PROPERTY)) {
            logUi.properties.set(SHOW_GIT_BRANCHES_LOG_PROPERTY, true)
          }
        }
      }
    branchesSearchFieldPanel.withBackground(UIUtil.getListBackground()).withBorder(createBorder(JBColor.border(), SideBorder.BOTTOM))
    branchesSearchFieldPanel.addToCenter(branchesSearchField)
    branchesTreePanel.addToTop(branchesSearchFieldPanel).addToCenter(branchesProgressStripe)
    branchesPanelExpandableController = ExpandablePanelController(toolbar.component, branchesButton, branchesTreePanel)
    branchViewSplitter.firstComponent = branchesTreePanel
    branchesTreeWithLogPanel.addToLeft(branchesPanelExpandableController.expandControlPanel).addToCenter(branchViewSplitter)
    mainPanel.isFocusCycleRoot = true
    mainPanel.focusTraversalPolicy = BRANCHES_UI_FOCUS_TRAVERSAL_POLICY
    startLoadingBranches()
    toggleBranchesPanelVisibility()
  }

  fun toggleBranchesPanelVisibility() {
    val showBranches = logUi.properties.get(SHOW_GIT_BRANCHES_LOG_PROPERTY)
    branchesPanelExpandableController.toggleExpand(showBranches)
  }

  private fun createFocusFilterFieldAction(searchField: Component) {
    DumbAwareAction.create { e ->
      val project = e.getRequiredData(CommonDataKeys.PROJECT)
      if (IdeFocusManager.getInstance(project).getFocusedDescendantFor(tree.component) != null) {
        IdeFocusManager.getInstance(project).requestFocus(searchField, true)
      }
      else {
        IdeFocusManager.getInstance(project).requestFocus(tree.component, true)
      }
    }.registerCustomShortcutSet(KeymapUtil.getActiveKeymapShortcuts("Find"), branchesTreePanel)
  }

  inner class BranchesTreePanel : BorderLayoutPanel(), DataProvider {
    override fun getData(dataId: String): Any? {
      if (GIT_BRANCHES.`is`(dataId)) {
        return tree.getSelectedBranches()
      }
      else if (VcsLogInternalDataKeys.LOG_UI_PROPERTIES.`is`(dataId)) {
        return logUi.properties
      }
      return null
    }
  }

  fun getMainComponent(): JComponent {
    return mainPanel
  }

  fun updateBranchesTree(initial: Boolean) {
    tree.update(initial)
  }

  fun refreshTree() {
    tree.refreshTree()
  }

  fun startLoadingBranches() {
    tree.component.emptyText.text = message("action.Git.Loading.Branches.progress")
    branchesTreePanel.isEnabled = false
    branchesProgressStripe.startLoading()
  }

  fun stopLoadingBranches() {
    tree.component.emptyText.text = getDefaultEmptyText()
    branchesTreePanel.isEnabled = true
    branchesProgressStripe.stopLoading()
  }

  override fun dispose() {
    disposeBranchesUi()
  }
}

internal class BranchesVcsLogUiFactory(logManager: VcsLogManager, logId: String, filters: VcsLogFilterCollection? = null)
  : BaseVcsLogUiFactory<BranchesVcsLogUi>(logId, filters, logManager.uiProperties, logManager.colorManager) {
  override fun createVcsLogUiImpl(logId: String,
                                  logData: VcsLogData,
                                  properties: MainVcsLogUiProperties,
                                  colorManager: VcsLogColorManagerImpl,
                                  refresher: VisiblePackRefresherImpl,
                                  filters: VcsLogFilterCollection?) =
    BranchesVcsLogUi(logId, logData, colorManager, properties, refresher, filters)
}

internal class BranchesVcsLogUi(id: String, logData: VcsLogData, colorManager: VcsLogColorManager,
                                uiProperties: MainVcsLogUiProperties, refresher: VisiblePackRefresher,
                                initialFilters: VcsLogFilterCollection?) :
  VcsLogUiImpl(id, logData, colorManager, uiProperties, refresher, initialFilters) {

  private val branchesUi =
    BranchesDashboardUi(logData.project, this)
      .also { branchesUi -> Disposer.register(this, branchesUi) }

  override fun createMainFrame(logData: VcsLogData, uiProperties: MainVcsLogUiProperties, filterUi: VcsLogFilterUiEx) =
    MainFrame(logData, this, uiProperties, filterUi, false)
      .apply {
        isFocusCycleRoot = false
        focusTraversalPolicy = null //new focus traversal policy will be configured include branches tree
        if (isDiffPreviewInEditor()) {
          VcsLogEditorDiffPreview(myProject, uiProperties, this)
        }
      }

  override fun getMainComponent() = branchesUi.getMainComponent()

  fun createDiffPreview(isInEditor: Boolean): VcsLogChangeProcessor {
    return mainFrame.createDiffPreview(isInEditor, mainFrame.changesBrowser)
  }
}

internal val SHOW_GIT_BRANCHES_LOG_PROPERTY =
  object : VcsLogProjectTabsProperties.CustomBooleanTabProperty("Show.Git.Branches") {
    override fun defaultValue(logId: String) = logId == MAIN_LOG_ID
  }

private class BranchViewSplitter(first: JComponent? = null, second: JComponent? = null)
  : OnePixelSplitter(false, "vcs.branch.view.splitter.proportion", 0.3f) {
  init {
    firstComponent = first
    secondComponent = second
  }
}

private class DiffPreviewSplitter(diffPreview: VcsLogChangeProcessor, uiProperties: VcsLogUiProperties, mainComponent: JComponent)
  : FrameDiffPreview<VcsLogChangeProcessor>(diffPreview, uiProperties, mainComponent,
                                            "vcs.branch.view.diff.splitter.proportion",
                                            Registry.`is`("vcs.log.diff.preview.vertical"), 0.3f) {
  override fun updatePreview(state: Boolean) {
    previewDiff.updatePreview(state)
  }
}

==> ideaIC-213.7172.25-sources.jar: git4idea/ui/branch/dashboard/BranchesDashboardUi.kt <==
// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package git4idea.ui.branch.dashboard

import com.intellij.dvcs.branch.GroupingKey
import com.intellij.icons.AllIcons
import com.intellij.ide.CommonActionsManager
import com.intellij.ide.DataManager
import com.intellij.ide.DefaultTreeExpander
import com.intellij.openapi.Disposable
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.ide.CopyPasteManager
import com.intellij.openapi.keymap.KeymapUtil
import com.intellij.openapi.progress.util.ProgressWindow
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.text.StringUtil
import com.intellij.openapi.vcs.changes.ui.ChangesBrowserBase
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.wm.IdeFocusManager
import com.intellij.ui.IdeBorderFactory.createBorder
import com.intellij.ui.JBColor
import com.intellij.ui.OnePixelSplitter
import com.intellij.ui.ScrollPaneFactory
import com.intellij.ui.SideBorder
import com.intellij.ui.components.panels.NonOpaquePanel
import com.intellij.ui.speedSearch.SpeedSearch
import com.intellij.util.concurrency.annotations.RequiresEdt
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.JBUI.Panels.simplePanel
import com.intellij.util.ui.StatusText.getDefaultEmptyText
import com.intellij.util.ui.UIUtil
import com.intellij.util.ui.components.BorderLayoutPanel
import com.intellij.util.ui.table.ComponentsListFocusTraversalPolicy
import com.intellij.vcs.log.VcsLogBranchLikeFilter
import com.intellij.vcs.log.VcsLogFilterCollection
import com.intellij.vcs.log.data.VcsLogData
import com.intellij.vcs.log.impl.MainVcsLogUiProperties
import com.intellij.vcs.log.impl.VcsLogApplicationSettings
import com.intellij.vcs.log.impl.VcsLogManager
import com.intellij.vcs.log.impl.VcsLogManager.BaseVcsLogUiFactory
import com.intellij.vcs.log.impl.VcsLogProjectTabsProperties
import com.intellij.vcs.log.impl.VcsLogProjectTabsProperties.MAIN_LOG_ID
import com.intellij.vcs.log.ui.VcsLogColorManager
import com.intellij.vcs.log.ui.VcsLogInternalDataKeys
import com.intellij.vcs.log.ui.VcsLogUiImpl
import com.intellij.vcs.log.ui.filter.VcsLogFilterUiEx
import com.intellij.vcs.log.ui.frame.MainFrame
import com.intellij.vcs.log.ui.frame.ProgressStripe
import com.intellij.vcs.log.util.VcsLogUtil
import com.intellij.vcs.log.visible.VisiblePackRefresher
import com.intellij.vcs.log.visible.VisiblePackRefresherImpl
import com.intellij.vcs.log.visible.filters.VcsLogFilterObject
import com.intellij.vcs.log.visible.filters.with
import com.intellij.vcs.log.visible.filters.without
import git4idea.i18n.GitBundle.message
import git4idea.i18n.GitBundleExtensions.messagePointer
import git4idea.repo.GitRepository
import git4idea.ui.branch.dashboard.BranchesDashboardActions.DeleteBranchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.FetchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.NewBranchAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ShowBranchDiffAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ShowMyBranchesAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.ToggleFavoriteAction
import git4idea.ui.branch.dashboard.BranchesDashboardActions.UpdateSelectedBranchAction
import org.jetbrains.annotations.ApiStatus
import java.awt.Component
import java.awt.datatransfer.DataFlavor
import java.awt.event.ActionEvent
import javax.swing.AbstractAction
import javax.swing.Action
import javax.swing.JComponent
import javax.swing.TransferHandler
import javax.swing.event.TreeSelectionListener

internal class BranchesDashboardUi(project: Project, private val logUi: BranchesVcsLogUi) : Disposable {
  private val uiController = BranchesDashboardController(project, this)

  private val tree = FilteringBranchesTree(project, BranchesTreeComponent(project), uiController, disposable = this)
  private val branchViewSplitter = BranchViewSplitter()
  private val branchesTreePanel = BranchesTreePanel().withBorder(createBorder(JBColor.border(), SideBorder.LEFT))
  private val branchesScrollPane = ScrollPaneFactory.createScrollPane(tree.component, true)
  private val branchesProgressStripe = ProgressStripe(branchesScrollPane, this, ProgressWindow.DEFAULT_PROGRESS_DIALOG_POSTPONE_TIME_MILLIS)
  private val branchesTreeWithLogPanel = simplePanel()
  private val mainPanel = simplePanel().apply { DataManager.registerDataProvider(this, uiController) }
  private val branchesSearchFieldPanel = simplePanel()
  private val branchesSearchField =
    NonOpaquePanel(tree.installSearchField().apply { textEditor.border = JBUI.Borders.emptyLeft(5) }).apply(UIUtil::setNotOpaqueRecursively)

  private lateinit var branchesPanelExpandableController: ExpandablePanelController

  private val treeSelectionListener = TreeSelectionListener {
    if (!branchesPanelExpandableController.isExpanded()) return@TreeSelectionListener

    val ui = logUi

    val properties = ui.properties

    if (properties[CHANGE_LOG_FILTER_ON_BRANCH_SELECTION_PROPERTY]) {
      updateLogBranchFilter()
    }
    else if (properties[NAVIGATE_LOG_TO_BRANCH_ON_BRANCH_SELECTION_PROPERTY]) {
      navigateToSelectedBranch(false)
    }
  }

  internal fun updateLogBranchFilter() {
    val ui = logUi
    val selectedFilters = tree.getSelectedBranchFilters()
    val oldFilters = ui.filterUi.filters
    val newFilters = if (selectedFilters.isNotEmpty()) {
      oldFilters.without(VcsLogBranchLikeFilter::class.java).with(VcsLogFilterObject.fromBranches(selectedFilters))
    } else {
      oldFilters.without(VcsLogBranchLikeFilter::class.java)
    }
    ui.filterUi.filters = newFilters
  }

  internal fun navigateToSelectedBranch(focus: Boolean) {
    val selectedReference = tree.getSelectedBranchFilters().singleOrNull() ?: return

    logUi.vcsLog.jumpToReference(selectedReference, focus)
  }

  internal fun toggleGrouping(key: GroupingKey, state: Boolean) {
    tree.toggleGrouping(key, state)
  }

  internal fun isGroupingEnabled(key: GroupingKey) = tree.isGroupingEnabled(key)

  internal fun getSelectedRepositories(branchInfo: BranchInfo): List<GitRepository> {
    return tree.getSelectedRepositories(branchInfo)
  }

  internal fun getSelectedRemotes(): Set<RemoteInfo> {
    return tree.getSelectedRemotes()
  }

  internal fun getRootsToFilter(): Set<VirtualFile> {
    val roots = logUi.logData.roots.toSet()
    if (roots.size == 1) return roots

    return VcsLogUtil.getAllVisibleRoots(roots, logUi.filterUi.filters)
  }

  private val BRANCHES_UI_FOCUS_TRAVERSAL_POLICY = object : ComponentsListFocusTraversalPolicy() {
    override fun getOrderedComponents(): List<Component> = listOf(tree.component, logUi.table,
                                                                  logUi.changesBrowser.preferredFocusedComponent,
                                                                  logUi.filterUi.textFilterComponent.textEditor)
  }

  private val showBranches get() = logUi.properties.get(SHOW_GIT_BRANCHES_LOG_PROPERTY)

  init {
    initMainUi()
    installLogUi()
    toggleBranchesPanelVisibility()
  }

  @RequiresEdt
  private fun installLogUi() {
    uiController.registerDataPackListener(logUi.logData)
    uiController.registerLogUiPropertiesListener(logUi.properties)
    uiController.registerLogUiFilterListener(logUi.filterUi)
    branchesSearchField.setVerticalSizeReferent(logUi.toolbar)
    branchViewSplitter.secondComponent = logUi.mainLogComponent
    mainPanel.add(branchesTreeWithLogPanel)
    tree.component.addTreeSelectionListener(treeSelectionListener)
  }

  @RequiresEdt
  private fun disposeBranchesUi() {
    branchViewSplitter.secondComponent.removeAll()
    uiController.removeDataPackListener(logUi.logData)
    uiController.removeLogUiPropertiesListener(logUi.properties)
    tree.component.removeTreeSelectionListener(treeSelectionListener)
  }

  private fun initMainUi() {
    val diffAction = ShowBranchDiffAction()
    diffAction.registerCustomShortcutSet(KeymapUtil.getActiveKeymapShortcuts("Diff.ShowDiff"), branchesTreeWithLogPanel)

    val deleteAction = DeleteBranchAction()
    val shortcuts = KeymapUtil.getActiveKeymapShortcuts("SafeDelete").shortcuts + KeymapUtil.getActiveKeymapShortcuts(
      "EditorDeleteToLineStart").shortcuts
    deleteAction.registerCustomShortcutSet(CustomShortcutSet(*shortcuts), branchesTreeWithLogPanel)

    createFocusFilterFieldAction(branchesSearchField)
    installPasteAction(tree)

    val toggleFavoriteAction = ToggleFavoriteAction()
    val fetchAction = FetchAction(this)
    val showMyBranchesAction = ShowMyBranchesAction(uiController)
    val newBranchAction = NewBranchAction()
    val updateSelectedAction = UpdateSelectedBranchAction()
    val defaultTreeExpander = DefaultTreeExpander(tree.component)
    val commonActionsManager = CommonActionsManager.getInstance()
    val expandAllAction = commonActionsManager.createExpandAllHeaderAction(defaultTreeExpander, branchesTreePanel)
    val collapseAllAction = commonActionsManager.createCollapseAllHeaderAction(defaultTreeExpander, branchesTreePanel)
    val actionManager = ActionManager.getInstance()
    val hideBranchesAction = actionManager.getAction("Git.Log.Hide.Branches")
    val settings = actionManager.getAction("Git.Log.Branches.Settings")

    val group = DefaultActionGroup()
    group.add(hideBranchesAction)
    group.add(Separator())
    group.add(newBranchAction)
    group.add(updateSelectedAction)
    group.add(deleteAction)
    group.add(diffAction)
    group.add(showMyBranchesAction)
    group.add(fetchAction)
    group.add(toggleFavoriteAction)
    group.add(actionManager.getAction("Git.Log.Branches.Navigate.Log.To.Selected.Branch"))
    group.add(Separator())
    group.add(settings)
    group.add(actionManager.getAction("Git.Log.Branches.Grouping.Settings"))
    group.add(expandAllAction)
    group.add(collapseAllAction)

    val toolbar = actionManager.createActionToolbar("Git.Log.Branches", group, false)
    toolbar.setTargetComponent(branchesTreePanel)

    val branchesButton = ExpandStripeButton(messagePointer("action.Git.Log.Show.Branches.text"), AllIcons.Actions.ArrowExpand)
      .apply {
        border = createBorder(JBColor.border(), SideBorder.RIGHT)
        addActionListener {
          if (logUi.properties.exists(SHOW_GIT_BRANCHES_LOG_PROPERTY)) {
            logUi.properties.set(SHOW_GIT_BRANCHES_LOG_PROPERTY, true)
          }
        }
      }
    branchesSearchFieldPanel.withBackground(UIUtil.getListBackground()).withBorder(createBorder(JBColor.border(), SideBorder.BOTTOM))
    branchesSearchFieldPanel.addToCenter(branchesSearchField)
    branchesTreePanel.addToTop(branchesSearchFieldPanel).addToCenter(branchesProgressStripe)
    branchesPanelExpandableController = ExpandablePanelController(toolbar.component, branchesButton, branchesTreePanel)
    branchViewSplitter.firstComponent = branchesTreePanel
    branchesTreeWithLogPanel.addToLeft(branchesPanelExpandableController.expandControlPanel).addToCenter(branchViewSplitter)
    mainPanel.isFocusCycleRoot = true
    mainPanel.focusTraversalPolicy = BRANCHES_UI_FOCUS_TRAVERSAL_POLICY
  }

  fun toggleBranchesPanelVisibility() {
    branchesPanelExpandableController.toggleExpand(showBranches)
    updateBranchesTree(true)
  }

  private fun createFocusFilterFieldAction(searchField: Component) {
    DumbAwareAction.create { e ->
      val project = e.getRequiredData(CommonDataKeys.PROJECT)
      if (IdeFocusManager.getInstance(project).getFocusedDescendantFor(tree.component) != null) {
        IdeFocusManager.getInstance(project).requestFocus(searchField, true)
      }
      else {
        IdeFocusManager.getInstance(project).requestFocus(tree.component, true)
      }
    }.registerCustomShortcutSet(KeymapUtil.getActiveKeymapShortcuts("Find"), branchesTreePanel)
  }

  private fun installPasteAction(tree: FilteringBranchesTree) {
    tree.component.actionMap.put(TransferHandler.getPasteAction().getValue(Action.NAME), object: AbstractAction () {
      override fun actionPerformed(e: ActionEvent?) {
        val speedSearch = tree.searchModel.speedSearch as? SpeedSearch ?: return
        val pasteContent =
          CopyPasteManager.getInstance().getContents<String>(DataFlavor.stringFlavor)
            // the same filtering logic as in javax.swing.text.PlainDocument.insertString (e.g. DnD to search field)
            ?.let { StringUtil.convertLineSeparators(it, " ") }
        speedSearch.type(pasteContent)
        speedSearch.update()
      }
    })
  }

  inner class BranchesTreePanel : BorderLayoutPanel(), DataProvider {
    override fun getData(dataId: String): Any? {
      return when {
        GIT_BRANCHES.`is`(dataId) -> tree.getSelectedBranches()
        GIT_BRANCH_FILTERS.`is`(dataId) -> tree.getSelectedBranchFilters()
        BRANCHES_UI_CONTROLLER.`is`(dataId) -> uiController
        VcsLogInternalDataKeys.LOG_UI_PROPERTIES.`is`(dataId) -> logUi.properties
        else -> null
      }
    }
  }

  fun getMainComponent(): JComponent {
    return mainPanel
  }

  fun updateBranchesTree(initial: Boolean) {
    if (showBranches) {
      tree.update(initial)
    }
  }

  fun refreshTree() {
    tree.refreshTree()
  }

  fun refreshTreeModel() {
    tree.refreshNodeDescriptorsModel()
  }

  fun startLoadingBranches() {
    tree.component.emptyText.text = message("action.Git.Loading.Branches.progress")
    branchesTreePanel.isEnabled = false
    branchesProgressStripe.startLoading()
  }

  fun stopLoadingBranches() {
    tree.component.emptyText.text = getDefaultEmptyText()
    branchesTreePanel.isEnabled = true
    branchesProgressStripe.stopLoading()
  }

  override fun dispose() {
    disposeBranchesUi()
  }
}

internal class BranchesVcsLogUiFactory(logManager: VcsLogManager, logId: String, filters: VcsLogFilterCollection? = null)
  : BaseVcsLogUiFactory<BranchesVcsLogUi>(logId, filters, logManager.uiProperties, logManager.colorManager) {
  override fun createVcsLogUiImpl(logId: String,
                                  logData: VcsLogData,
                                  properties: MainVcsLogUiProperties,
                                  colorManager: VcsLogColorManager,
                                  refresher: VisiblePackRefresherImpl,
                                  filters: VcsLogFilterCollection?) =
    BranchesVcsLogUi(logId, logData, colorManager, properties, refresher, filters)
}

internal class BranchesVcsLogUi(id: String, logData: VcsLogData, colorManager: VcsLogColorManager,
                                uiProperties: MainVcsLogUiProperties, refresher: VisiblePackRefresher,
                                initialFilters: VcsLogFilterCollection?) :
  VcsLogUiImpl(id, logData, colorManager, uiProperties, refresher, initialFilters) {

  private val branchesUi =
    BranchesDashboardUi(logData.project, this)
      .also { branchesUi -> Disposer.register(this, branchesUi) }

  internal val mainLogComponent: JComponent
    get() = mainFrame

  internal val changesBrowser: ChangesBrowserBase
    get() = mainFrame.changesBrowser

  override fun createMainFrame(logData: VcsLogData, uiProperties: MainVcsLogUiProperties,
                               filterUi: VcsLogFilterUiEx, isEditorDiffPreview: Boolean) =
    MainFrame(logData, this, uiProperties, filterUi, isEditorDiffPreview, this)
      .apply {
        isFocusCycleRoot = false
        focusTraversalPolicy = null //new focus traversal policy will be configured include branches tree
      }

  override fun getMainComponent() = branchesUi.getMainComponent()
}

@ApiStatus.Internal
val SHOW_GIT_BRANCHES_LOG_PROPERTY =
  object : VcsLogProjectTabsProperties.CustomBooleanTabProperty("Show.Git.Branches") {
    override fun defaultValue(logId: String) = logId == MAIN_LOG_ID
  }

@ApiStatus.Internal
val CHANGE_LOG_FILTER_ON_BRANCH_SELECTION_PROPERTY =
  object : VcsLogApplicationSettings.CustomBooleanProperty("Change.Log.Filter.on.Branch.Selection") {
    override fun defaultValue() = false
  }

@ApiStatus.Internal
val NAVIGATE_LOG_TO_BRANCH_ON_BRANCH_SELECTION_PROPERTY =
  object : VcsLogApplicationSettings.CustomBooleanProperty("Navigate.Log.To.Branch.on.Branch.Selection") {
    override fun defaultValue() = false
  }

private class BranchViewSplitter(first: JComponent? = null, second: JComponent? = null)
  : OnePixelSplitter(false, "vcs.branch.view.splitter.proportion", 0.3f) {
  init {
    firstComponent = first
    secondComponent = second
  }
}
